"""Manages all path metadata."""
from pathlib import Path
from datetime import datetime

##################
# Executor paths #
##################
EXEC_R_SCRIPT_PATH = Path('/share/singularity-images/lbd/shells/singR.sh')
R_SINGULARITY_IMAGE_PATH = Path('/ihme/singularity-images/lbd/releases/lbd_full_20200128.simg')

################
# Shared paths #
################
# Concrete share drive root
COVID_19 = Path('/ihme/covid-19')
ARCHIVE_ROOT = COVID_19 / 'archive'

# Shared config for running rclone on the IHME OneDrive
RCLONE_CONFIG_PATH = COVID_19 / '.config' / 'rclone' / 'rclone.conf'

# Top level directories.  These represent, mostly, outputs of pipeline stages.
# Keep them in alphabetical order.
# TODO: This should definitely be in a container.
AGE_SPECIFIC_RATES_ROOT = COVID_19 / 'age-specific-rates'
AIRLINE_DATA_ROOT = COVID_19 / 'airline-data'
DATA_FIXES_ROOT = COVID_19 / 'data-fixes'
DATA_INTAKE_ROOT = COVID_19 / 'data_intake'
DURATIONS_ROOT = COVID_19 / 'durations'
EXCESS_MORTALITY_ROOT = COVID_19 / 'excess-mortality'
GBD_ROOT = COVID_19 / 'gbd'
HISTORICAL_MODEL_ROOT = COVID_19 / 'historical-model'
MANDATES_ROOT = COVID_19 / 'mandates'
MASK_USE_OUTPUT_ROOT = COVID_19 / 'mask-use-outputs'
MOBILITY_COVARIATES_OUTPUT_ROOT = COVID_19 / 'mobility-covariate'
MOBILITY_COVARIATES_GPR_OUTPUT_ROOT = MOBILITY_COVARIATES_OUTPUT_ROOT / 'gpr_outputs'
MODEL_COMPARISON_INPUTS_ROOT = COVID_19 / 'model-comparison-inputs'
MODEL_INPUTS_ROOT = COVID_19 / 'model-inputs'
MORTALITY_AGE_PATTERN_ROOT = COVID_19 / 'mortality-age-pattern'
MORTALITY_SCALARS_ROOT = COVID_19 / 'mortality-scalars'
PAST_INFECTIONS_ROOT = COVID_19 / 'past-infections'
PNEUMONIA_OUTPUT_ROOT = COVID_19 / 'pneumonia'
POPULATION_DENSITY_OUTPUT_ROOT = COVID_19 / 'population-density'
RESULTRON_ROOT = COVID_19 / 'resultron'
SEIR_COUNTERFACTUAL_INPUT_ROOT = COVID_19 / 'seir-counterfactual-input'
SEIR_COUNTERFACTUAL_ROOT = COVID_19 / 'seir-counterfactual'
SEIR_COVARIATE_PRIORS_ROOT = COVID_19 / 'seir-covariate-priors'
SEIR_COVARIATES_OUTPUT_ROOT = COVID_19 / 'seir-covariates'
SEIR_DIAGNOSTICS_OUTPUTS = COVID_19 / 'seir-diagnostics'
SEIR_FIT_ROOT = COVID_19 / 'seir-fit'
SEIR_FORECAST_OUTPUTS = COVID_19 / 'seir-forecast'
SEIR_OOS_ANALYSIS_OUTPUTS = COVID_19 / 'seir-oos-analysis'
SEIR_FINAL_OUTPUTS = COVID_19 / 'seir-outputs'
SEIR_PREPROCESS_ROOT = COVID_19 / 'seir-preprocess'
SEIR_REGRESSION_OUTPUTS = COVID_19 / 'seir-regression'
SHAPEFILE_ROOT = COVID_19 / 'shapefiles'
SNAPSHOT_ROOT = COVID_19 / 'snapshot-data'
STATIC_DATA_INPUTS_ROOT = COVID_19 / 'static-data'
SYMPTOM_SURVEY_ROOT = COVID_19 / 'symptom-survey-data'
TESTING_OUTPUT_ROOT = COVID_19 / 'testing-outputs'
UNVERSIONED_INPUTS_ROOT = COVID_19 / 'unversioned-inputs'
VACCINE_COVERAGE_OUTPUT_ROOT = COVID_19 / 'vaccine-coverage'
VACCINE_EFFICACY_ROOT = COVID_19 / 'vaccine-efficacy'
VARIANT_OUTPUT_ROOT = COVID_19 / 'variant-scaleup'
VISIT_VOLUME_ROOT = COVID_19 / 'visit-volume/'
WANING_IMMUNITY_OUTPUT_ROOT = COVID_19 / 'waning-immunity'
WEBSCRAPER_ROOT = COVID_19 / 'webscrape'


######################################
# Shared file and subdirectory names #
######################################
METADATA_FILE_NAME = Path('metadata.yaml')

LOG_DIR = Path("logs")
LOG_FILE_NAME = Path("master_log.txt")
DETAILED_LOG_FILE_NAME = Path("master_log.json")

BEST_LINK = Path('best')
LATEST_LINK = Path('latest')
PRODUCTION_RUN = Path('production-runs')


def latest_production_snapshot_path():
    return _latest_prod_path(SNAPSHOT_ROOT)


def latest_production_etl_path():
    return _latest_prod_path(MODEL_INPUTS_ROOT)


def _latest_prod_path(prefix: Path):
    prod_run_dir = prefix / PRODUCTION_RUN
    prod_runs = [d for d in prod_run_dir.iterdir()]
    sorted_runs = list(sorted(prod_runs, key=lambda p: datetime.strptime(p.stem, '%Y_%m_%d')))
    return sorted_runs[-1]


#################
# I/O utilities #
#################

# TODO: Set the ihme-covid user group.  Don't think I can do this without sudo.
DIRECTORY_PERMISSIONS = 0o775
FILE_PERMISSIONS = 0o664


def make_dir_tree(directory: Path):
    """Makes all directories and their parents with the correct permissions.

    Parameters
    ----------
    directory
        The directory to make.

    This skirts around the default behavior of :func:`Path.mkdir` which
    mimics `mkdir -p` which will create a dir with requested permissions,
    but all parent directories with default permissions.

    """
    to_create = []
    p = directory
    while not p.exists():
        to_create.append(p)
        p = p.parent
    while to_create:
        to_create.pop().mkdir(DIRECTORY_PERMISSIONS)


def recursive_set_permissions(path: Path):
    """Recursively set permissions to defaults."""
    if path.is_file():
        path.chmod(FILE_PERMISSIONS)
    else:
        path.chmod(DIRECTORY_PERMISSIONS)
        for p in path.iterdir():
            recursive_set_permissions(p)



from covid_shared.workflow.specification import (
    TaskSpecification,
    WorkflowSpecification,
)
from covid_shared.workflow.template import (
    TaskTemplate,
    WorkflowTemplate,
)
from covid_shared.workflow.utilities import (
    get_jobmon_tool,
)



"""Wrappers for IHME specific dependencies.

This module explicitly declares and wraps IHME specific code bases
to prevent CI failures at import time.

"""
import importlib
import sys
from pathlib import Path

import pandas as pd


def _lazy_import_callable(module_path: str, object_name: str):
    """Delays import errors for callables until called."""
    try:
        module = importlib.import_module(module_path)
        return getattr(module, object_name)
    except ModuleNotFoundError:
        def f(*args, **kwargs):
            raise ModuleNotFoundError(
                f"No module named '{module_path}' and so we cannot find '{object_name}'. "
                f"Ensure you have a file in your home directory at '~/.pip/pip.conf' with contents\n\n"
                f"[global]\n"
                f"extra-index-url = https://artifactory.ihme.washington.edu/artifactory/api/pypi/pypi-shared/simple\n"
                f"trusted-host = artifactory.ihme.washington.edu/artifactory/api/pypi/pypi-shared\n\n"
                f"and run 'pip install {module_path.split('.')[0]}'")
        return f


try:
    from db_queries.api.internal import get_location_hierarchy_by_version
except ModuleNotFoundError:
    get_location_hierarchy_by_version = _lazy_import_callable('db_queries.api.internal',
                                                              'get_location_hierarchy_by_version')


def load_location_hierarchy(location_set_version_id: int = None,
                            location_file: Path = None):
    assert ((location_set_version_id and not location_file)
            or (not location_set_version_id and location_file))

    if location_set_version_id:
        return get_location_hierarchy_by_version(
            location_set_version_id=location_set_version_id,
        )
    else:
        return pd.read_csv(location_file)

##############
# GROSS HACK #
##############
# jobmon_uge uses structlog, a library which does an endrun around
# python standard logging and then dumps a bunch of useless information
# to stdout. Before we import structlog (via jobmon_uge via jobmon),
# put the name in the system's list of imported modules as an alias
# to the python std library logging module.
sys.modules['structlog'] = sys.modules['logging']

try:
    from jobmon.client.api import Tool
    from jobmon.client.task import Task
    from jobmon.client.workflow import WorkflowRunStatus
    from jobmon.exceptions import WorkflowAlreadyComplete
except ModuleNotFoundError:
    Tool = _lazy_import_callable('jobmon.client.api', 'Tool')
    Task = _lazy_import_callable('jobmon.client.task', 'Task')
    WorkflowRunStatus = _lazy_import_callable('jobmon.client.workflow', 'WorkflowRunStatus')
    WorkflowAlreadyComplete = _lazy_import_callable('jobmon.exceptions', 'WorkflowAlreadyComplete')



from pathlib import Path
from typing import Optional

from covid_shared.cli_tools.metadata import Metadata, RunMetadata
from covid_shared.cli_tools.run_directory import make_links


def finish_application(
    run_metadata: RunMetadata,
    app_metadata: Metadata,
    run_directory: Path,
    mark_as_best: bool,
    production_tag: Optional[str],
    quick: Optional[int] = None
) -> None:
    """
    Every cli tool should do the following:
        1. serialize metadata to disk
        2. update symlinks (if successful application)
        3. raise exception (if not sucessful)

    Parameters
    ------------
        run_metadata: metadata from cli invocation
        app_metadata: metadata from application getting monitored
        run_directory: output path of application
        mark_as_best: whether to update 'best' symlink
        production_tag: what string to tag prod run, if any
    """
    run_metadata['app_metadata'] = app_metadata.to_dict()
    run_metadata.dump(run_directory / 'metadata.yaml')

    # Configure latest or best symlink
    make_links(app_metadata, run_directory, mark_as_best, production_tag, quick)

    _raise_if_exception(app_metadata)


def _raise_if_exception(app_metadata: Metadata) -> None:
    """
    CLI tools should return a non-zero error code in the case of an error.
    cli_tools.monitor_application catches exceptions so we need to look for
    and raise any exception in the app metadata if it exists.

    There are 3 cases:
        1. no exception. In that case metadata contains success=True
        2. user interrupt. In that case metadata contains success=False and
            error_info='User interrupt.'
        3. any other exception: metadata contains success=False and error_info
            is a dictionary with keys 'exception_type', 'exception_value',
            'exc_traceback'
    """
    if app_metadata['success']:
        return

    error_info = app_metadata['error_info']
    user_interrupt = 'interrupt' in error_info
    if user_interrupt:
        raise KeyboardInterrupt
    else:
        raise error_info['exception_value']



import logging
from pathlib import Path
import sys
from typing import TextIO, Tuple, Union

from loguru import logger

from covid_shared import paths
from covid_shared.shell_tools import mkdir


JOBMON_LOGGING_LEVEL = 5  # A lower level than logging.DEBUG.
DEFAULT_LOG_MESSAGING_FORMAT = (
    '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
    '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> '
    '- <level>{message}</level> - {extra}'
)
LOG_FORMATS = {
    # Keys are verbosity.  Specify special log formats here.
    0: ("WARNING", DEFAULT_LOG_MESSAGING_FORMAT),
    1: ("INFO", DEFAULT_LOG_MESSAGING_FORMAT),
    2: ("DEBUG", DEFAULT_LOG_MESSAGING_FORMAT),
    3: (JOBMON_LOGGING_LEVEL, DEFAULT_LOG_MESSAGING_FORMAT),
}


def configure_logging_to_terminal(verbose: int) -> None:
    """Setup logging to sys.stdout.

    This is presumed to be one of the first calls made in an
    application entry point. Any logging that occurs before this
    call won't be intercepted or handled with the standard
    logging configuration.

    """
    logger.remove(0)  # Clear default configuration
    level = LOG_FORMATS.get(verbose, LOG_FORMATS[max(LOG_FORMATS.keys())])[0]
    intercept_jobmon_logs(level)
    add_logging_sink(sys.stdout, verbose, colorize=True)


def configure_logging_to_files(output_path: Path) -> None:
    """Sets up logging to a file in an output directory.

    Logs to files are done with the highest verbosity to allow
    for debugging if necessary.

    """
    log_path = output_path / paths.LOG_DIR
    mkdir(log_path, exists_ok=True)
    add_logging_sink(
        output_path / paths.LOG_DIR / paths.DETAILED_LOG_FILE_NAME,
        verbose=3,
        serialize=True,
    )
    add_logging_sink(
        output_path / paths.LOG_DIR / paths.LOG_FILE_NAME,
        verbose=3,
    )


def add_logging_sink(sink: Union[TextIO, Path],
                     verbose: int,
                     colorize: bool = False,
                     serialize: bool = False) -> None:
    """Add a new output file handle for logging."""
    level, message_format = LOG_FORMATS.get(
        verbose, LOG_FORMATS[max(LOG_FORMATS.keys())]
    )
    logger.add(
        sink,
        colorize=colorize,
        level=level,
        format=message_format,
        serialize=serialize,
        filter={
            # Suppress logs up to the level provided.
            'urllib3': 'WARNING',  # Uselessly (for us) noisy.
        }
    )


class JobmonInterceptHandler(logging.Handler):
    """Intercepts and handles logging messages from jobmon.

    Jobmon logs a lot of information on its own set of loggers at
    sometimes unexpected priority levels. Many of these messages
    are not useful in an application as they generally deal with
    jobmon internals. This handler does both filtering and parsing
    of jobmon log messages in order to suppress uninteresting
    messages and to summarize status information.

    The original messages are still emitted at the
    JOBMON_LOGGING_LEVEL for debugging purposes.

    As the core loop of jobmon is multithreaded, there is technically
    a race condition here as we could be asked to process a second
    message before we've finished the first. In practice, the
    processing here is so fast and the core jobmon loop sleeps
    so long that we'll never have an issue.

    """
    # log handling strategies
    _FILTER = 0
    _LIFT = 1
    _RE_EMIT = 2

    def emit(self, record: logging.LogRecord) -> None:
        # Grab some metadata about the record.
        level, depth = self._get_level_and_depth(record)

        # Parse/filter the message. Filtered messages
        message, log_strategy = self._get_message_and_strategy(record)

        if log_strategy == self._FILTER:
            pass
        elif log_strategy == self._LIFT:
            logger.info(message)
        else:  # log_strategy == self._RE_EMIT
            logger.opt(depth=depth, exception=record.exc_info).log(level, message)

        # Finally, relog the original at the jobmon logging level.
        original_message = record.getMessage()
        logger.opt(depth=depth, exception=record.exc_info).log(
            JOBMON_LOGGING_LEVEL, original_message
        )

    def _get_message_and_strategy(self, record: logging.LogRecord) -> Tuple[str, int]:
        # If it's more important than an INFO message just re-emit.
        if record.levelno > logging.INFO:
            return record.getMessage(), self._RE_EMIT
        elif self._should_drop(record):
            return record.getMessage(), self._FILTER
        else:
            return self._parse_record(record)

    def _should_drop(self, record: logging.LogRecord) -> bool:
        # drop_list is tuples of (logger_name, [ignore_functions])
        # where logger_name is usually the name of the module
        # being logged from (though occasionally a module
        # higher in the call stack) and each ignore_function is
        # the function or method that is logging a message.
        drop_list = (
            (
                'jobmon.client.workflow',
                ['add_task', '_distributor_alive']
            ),
            (
                'jobmon.client.distributor.distributor_service',
                ['_create_task_instance', '_keep_distributing', 'heartbeat',
                 '_get_lost_task_instances', '_distribute_forever']
            ),
            # Finally, drop any messages from these functions
            # regardless of where they're called from.
            (record.name, ['_send_request', 'is_5XX'])
        )
        for logger_name, ignore_functions in drop_list:
            if record.name == logger_name and record.funcName in ignore_functions:
                return True
        return False

    def _parse_record(self, record: logging.LogRecord) -> Tuple[str, int]:
        msg = record.getMessage()
        strategy = self._RE_EMIT

        if record.name == 'jobmon.client.distributor.distributor_service':
            # Couldn't filter these before without looking in the message.
            strategy = self._FILTER
            # We don't want these messages, but we want what's in them.
            if 'active distributor_ids:' in msg:
                queued_or_running = str(len(msg.split(',')))
                # This updates every 30s, which is a reasonable
                # frequency to log the workflow status.
                msg = f'Queued or running: {queued_or_running}'
                strategy = self._LIFT

        return msg, strategy

    @staticmethod
    def _get_level_and_depth(record: logging.LogRecord) -> Tuple[Union[int, str], int]:
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        return level, depth


def intercept_jobmon_logs(level: Union[str, int]) -> None:
    """Add a handler to all the jobmon loggers to intercept and filter/parse the messages."""
    handler = JobmonInterceptHandler(level)
    for name, logger_ in logging.Logger.manager.loggerDict.items():
        if 'jobmon' in name and not isinstance(logger_, logging.PlaceHolder):
            logger_.handlers = [handler]


def list_loggers():
    """Utility function for analyzing the logging environment."""
    root_logger = logging.getLogger()
    print("Root logger: ", root_logger)
    for h in root_logger.handlers:
        print(f'     %s' % h)

    print("Other loggers")
    print("=============")
    for name, logger_ in logging.Logger.manager.loggerDict.items():
        print('+ [%-20s] %s ' % (name, logger_))
        if not isinstance(logger_, logging.PlaceHolder):
            for h in logger_.handlers:
                print('     %s' % h)



import datetime
from pathlib import Path
import re
from typing import Optional, Union, Tuple
from warnings import warn

from loguru import logger

from covid_shared import paths
from covid_shared.cli_tools.metadata import Metadata
from covid_shared.shell_tools import mkdir


def make_run_directory(output_root: Union[str, Path]) -> Path:
    """Convenience function for making a new run directory and getting its path."""
    run_directory = get_run_directory(output_root)
    mkdir(run_directory)
    return run_directory


def get_run_directory(output_root: Union[str, Path]) -> Path:
    """Gets a path to a datetime directory for a new output.

    Parameters
    ----------
    output_root
        The root directory for all outputs.

    """
    output_root = Path(output_root).resolve()
    launch_time = datetime.datetime.now().strftime("%Y_%m_%d")
    today_runs = [int(run_dir.name.split('.')[1]) for run_dir in output_root.iterdir() if
                  run_dir.name.startswith(launch_time)]
    run_version = max(today_runs) + 1 if today_runs else 1
    datetime_dir = output_root / f'{launch_time}.{run_version:0>2}'
    return datetime_dir


def get_last_stage_directory(last_stage_version: Union[str, Path], last_stage_directory: Union[str, Path] = None,
                             last_stage_root: Path = None) -> Path:
    """Get the directory containing the results of the last pipeline stage.

    Parameters
    ----------
    last_stage_version
        The path to the version. Can either be an absolute path, or a path relative to the last_stage_root

    last_stage_directory
        Deprecated parameter. The path to the version. Should use last_stage_version instead

    last_stage_root
        The path to the root of the resource. Must be an absolute path.
    """
    if last_stage_directory:
        warn(f'Usage of the "last_stage_directory" argument is deprecated. Please use "last_stage_version instead.',
             Warning)
    last_stage_directory = last_stage_directory if last_stage_directory is not None else last_stage_version
    # If last_stage_directory is an absolute path, the last_stage_root will be ignored here
    last_stage_directory = (last_stage_root / last_stage_directory if last_stage_root is not None
                            else last_stage_directory)
    if not last_stage_directory.is_absolute():
        raise ValueError(f'Invalid version path: {last_stage_directory}')
    return last_stage_directory


def setup_directory_structure(output_root: Union[str, Path], with_production: bool = False) -> None:
    """Sets up a best and latest directory for results versioning.

    Parameters
    ----------
    output_root
        The root directory for all outputs.
    with_production
        If true, additionally sets up a `production-run` sub-directory within
        the primary output root.

    """
    mkdir(output_root, exists_ok=True, parents=True)
    output_root = Path(output_root).resolve()
    for link in [paths.BEST_LINK, paths.LATEST_LINK]:
        link_path = output_root / link
        if not link_path.is_symlink() and not link_path.exists():
            mkdir(link_path)

    if with_production:
        production_dir = output_root / paths.PRODUCTION_RUN
        mkdir(production_dir, exists_ok=True)


def make_links(app_metadata: Metadata, run_directory: Path,
               mark_as_best: bool, production_tag: Optional[str], quick: Optional[int] = None) -> None:
    if app_metadata['success'] and not quick:
        mark_latest(run_directory)

        if mark_as_best:
            mark_best(run_directory)

            if production_tag:
                try:
                    mark_production(run_directory, production_tag)
                except ValueError:
                    logger.warning(f'Invalid production tag {production_tag}. Run not marked for production.'
                                f'Please provide production tag in format YYYY_MM_DD.')


def mark_best(run_directory: Union[str, Path]) -> None:
    """Marks an output directory as the best source of data."""
    run_directory = Path(run_directory).resolve()
    mark_best_explicit(run_directory, run_directory.parent)


def mark_latest(run_directory: Union[str, Path]) -> None:
    """Marks an output directory as the latest source of data."""
    run_directory = Path(run_directory).resolve()
    mark_latest_explicit(run_directory, run_directory.parent)


def mark_production(run_directory: Union[str, Path], date: str = None) -> None:
    """Marks a run as a production run.

    Raises
    ------
    ValueError
        If date is not in the format YYYY_MM_DD

    """
    run_directory = Path(run_directory).resolve()
    mark_production_explicit(run_directory, run_directory.parent / paths.PRODUCTION_RUN, date)


def mark_production_explicit(run_directory: Union[str, Path], version_root: Union[str, Path], date: str = None) -> None:
    """Marks a run as a production run within a specific version root.

    Raises
    ------
    ValueError
        If date is not in the format YYYY_MM_DD

    """
    if date is None:
        date = datetime.datetime.now().strftime('%Y_%m_%d')
    else:
        # Will raise a Value error if not the correct format.
        datetime.datetime.strptime(date, '%Y_%m_%d')
    mark_explicit(run_directory, version_root, date)


def mark_best_explicit(run_directory: Union[str, Path], version_root: Union[str, Path]) -> None:
    """Marks a directory best within a specific version root."""
    mark_explicit(run_directory, version_root, paths.BEST_LINK)


def mark_latest_explicit(run_directory: Union[str, Path], version_root: Union[str, Path]) -> None:
    """Marks a directory latest within a specific version root."""
    mark_explicit(run_directory, version_root, paths.LATEST_LINK)


def mark_explicit(run_directory: Union[str, Path], version_root: Union[str, Path], link_name: Union[str, Path]) -> None:
    """Makes or moves a link name to the run directory in a version root."""
    run_directory = Path(run_directory).resolve()
    version_root = Path(version_root).resolve()
    link_file = version_root / link_name
    move_link(link_file, run_directory)


def move_link(symlink_file: Path, link_target: Path) -> None:
    """Removes an old symlink and links it to something else."""
    if symlink_file.is_symlink():
        symlink_file.unlink()
    elif symlink_file.is_dir():  # We have set this up be a directory at the start.
        symlink_file.rmdir()
    elif symlink_file.exists():  # A file exists but isn't a symlink or a directory
        raise ValueError(f'{str(symlink_file)} is not a symlink or a directory')
    symlink_file.symlink_to(link_target, target_is_directory=True)


def get_current_previous_version(current_run_directory: Path, previous_run_directory: Path = None,
                                 resolved_name: bool = True) -> Tuple[str, str]:
    """
    Takes a full path to the current version, and returns the string name of the current and previous versions.

    Parameters
    ----------
    current_run_directory
        Path of the current version.
    previous_run_directory
        Path of a previous version. If unspecified, the previous version is taken to be the latest run from a date
        prior to the current run's date
    resolved_name
        If False, resolves a potential symlink to a canonical location

    """
    run_matcher = r'^\d{4}'

    parent_dir = current_run_directory.parent
    current_version = current_run_directory.name
    current_version_resolved = current_run_directory.resolve().name
    current_version_date = current_version_resolved.split('.')[0]
    previous_version = previous_run_directory.name if previous_run_directory else sorted([
        p.name for p in parent_dir.iterdir() if re.search(run_matcher, p.name) and current_version_date > p.name
    ])[-1]
    return current_version_resolved if resolved_name else current_version, previous_version



__all__ = [
    "__title__", "__summary__", "__uri__", "__author__", "__email__", "__version__",
]

__title__ = "covid-shared"
__summary__ = "Shared utilities for the covid model pipeline."
__uri__ = "https://github.com/ihmeuw/covid-shared"

__version__ = '2.0.0'

__author__ = "James Collins"
__email__ = "collijk@uw.edu"



def validate_best_and_production_tags(modifier: bool, mark_best: bool, production_tag: str):
    """
    Incomplete runs should never be marked 'best' or for production, eg quick, trial-run, etc
    """
    if modifier and (mark_best or production_tag):
        raise ValueError("Cannot mark an incomplete run as best or for production.")



"""Primitives for construction jobmon workflows."""
import abc
import functools
from pathlib import Path
from typing import Dict, Type, TypeVar

from loguru import logger

from covid_shared import paths
from covid_shared.ihme_deps import (
    Task,
    WorkflowRunStatus,
)
from covid_shared.workflow.specification import (
    TaskSpecification,
    WorkflowSpecification,
)
from covid_shared.workflow.utilities import (
    JobmonTool,
    make_log_dirs,
    get_cluster_name,
)


class TaskTemplate(abc.ABC):
    """Factory class for a parameterized task.

    Subclasses are intended to inherit and provide string templates for the
    class variables ``task_name_template`` which will construct cluster
    job names from the task args and ``command_template`` which will
    resolve the task args into a job executable by bash.

    """

    task_name_template: str
    command_template: str
    node_args: list
    task_args: list
    tool: JobmonTool

    def __init__(self, name: str, task_specification: TaskSpecification):
        self.jobmon_template = self.tool.get_task_template(
            template_name=name,
            command_template=self.command_template,
            node_args=self.node_args,
            task_args=self.task_args,
        )
        self.params = task_specification.to_dict()

    def get_task(self, *_, **kwargs) -> Task:
        """Resolve job arguments into a bash executable task for jobmon."""
        task = self.jobmon_template.create_task(
            compute_resources=self.params,
            name=self.task_name_template.format(**kwargs),
            max_attempts=1,
            **kwargs,
        )
        return task


TTaskTemplate = TypeVar('TTaskTemplate', bound=TaskTemplate)


class WorkflowTemplate(abc.ABC):
    """Factory for building and running workflows from specifications.

    Subclasses are intended to inherit and provide a string template for the
    class variable ``workflow_name_template`` which takes an output version
    string and maps it to a unique workflow name, and a dictionary mapping
    task type names to concrete subclasses of the ``TaskTemplate`` base
    class as the ``task_template_classes`` class variable. This mapping
    is tightly coupled with the ``tasks`` mapping of the associated
    workflow specification (ie, they should have exactly the same keys).

    When instantiated, the ``WorkflowTemplate`` subclasses will
    produce a jobmon workflow from the provided workflow specification,
    set up output logging directories, and produce concrete templates for
    workflow tasks.

    Subclass implementers must provide an implementation of ``attach_tasks``
    which takes relevant model parameters as arguments and builds and attaches
    an appropriate task dag to the jobmon workflow.

    """

    tool: JobmonTool
    workflow_name_template: str = None
    task_template_classes: Dict[str, Type[TTaskTemplate]]
    fail_fast: bool = True

    def __init__(self, version: str, workflow_specification: WorkflowSpecification):
        self.version = version
        assert workflow_specification.tasks.keys() == self.task_template_classes.keys()
        self.task_templates = self.build_task_templates(workflow_specification.task_specifications)

        stdout, stderr = make_log_dirs(Path(version) / paths.LOG_DIR)

        cluster = get_cluster_name()
        if cluster == 'slurm' and workflow_specification.queue == 'd.q':
            raise RuntimeError(
                "The d.q is unavailable on the SLURM cluster. Please update the "
                "'workflow.queue' key in your specification and retry.")

        resources = {
            'stdout': stdout,
            'stderr': stderr,
            'project': workflow_specification.project,
        }

        self.workflow = self.tool.create_workflow(
            name=self.workflow_name_template.format(version=version),
            default_cluster_name=cluster,
            default_compute_resources_set={
                cluster: resources,
            }
        )

        ##############
        # GROSS HACK #
        ##############
        # workflow.run no longer gives back a workflow status, which means
        # we can't get the workflow run id to log on a failure. This patches
        # one of the internal methods that generates a workflow run to hang
        # onto the workflow run id as an attribute of the workflow.
        old_create_workflow_run = self.workflow._create_workflow_run

        @functools.wraps(old_create_workflow_run)
        def _my_create_workflow_run(*args, **kwargs):
            # Call __func__ so we don't get two copies of self.
            client_wfr = old_create_workflow_run.__func__(*args, **kwargs)
            self.workflow.workflow_run_id = client_wfr.workflow_run_id
            return client_wfr

        self._monkey_patch_method(
            original_method=old_create_workflow_run,
            new_method=_my_create_workflow_run,
        )

    @staticmethod
    def _monkey_patch_method(original_method, new_method):
        # Invoke the descriptor protocol to bind the wrapped method to the
        # component instance.
        rebound_method = new_method.__get__(
            original_method.__self__,
            original_method.__self__.__class__,
        )
        # Then update the instance dictionary to reflect that the wrapped
        # method is bound to the original name.
        setattr(
            original_method.__self__,
            original_method.__name__,
            rebound_method
        )

    def build_task_templates(self, task_specifications: Dict[str, TaskSpecification]) -> Dict[str, TaskTemplate]:
        """Parses task specifications into task templates."""
        task_templates = {}
        for task_name, task_specification in task_specifications.items():
            task_templates[task_name] = self.task_template_classes[task_name](task_name, task_specification)
        return task_templates

    @abc.abstractmethod
    def attach_tasks(self, *args, **kwargs) -> None:
        """Turn model arguments into jobmon workflow tasks."""
        pass

    def run(self) -> None:
        """Execute the constructed workflow."""
        try:
            r = self.workflow.run(
                fail_fast=self.fail_fast,
                seconds_until_timeout=60*60*24,
            )
        except RuntimeError:
            # fail_fast now induces a runtime error instead of returning an error
            # status, which is unexpected behavior.  Patch around this for now.
            r = WorkflowRunStatus.ERROR
        if r != WorkflowRunStatus.DONE:
            raise RuntimeError(
                f'Workflow failed with status {r}.\n'
                f'Workflow run id: {self.workflow.workflow_run_id}.'
            )



from pathlib import Path
import socket
from typing import Tuple, Union

from covid_shared import shell_tools
from covid_shared.ihme_deps import Tool


class JobmonTool:
    """Lazy initialization of the jobmon Tool.

    The Task and Workflow templates both create a jobmon tool at import time.
    This has the unfortunate side effect of talking to external services
    before anything else happens. This wrapper class delays initialization
    of the tool (and therefore the service communication) until runtime.

    """

    def __init__(self, package):
        if not hasattr(package, '__jobmon_tool_version__'):
            raise AttributeError(
                f"Package {package.__name__} must define a __jobmon_tool_version__ attribute "
                f"in its package level namespace.")
        else:
            self._package_name = package.__name__
            self._active_tool_version_id = package.__jobmon_tool_version__
        self._tool = None

    def lazy_init_tool(self):
        if self._tool is None:
            self._tool = Tool(self._package_name)
            self._tool.active_tool_version_id = self._active_tool_version_id

    def create_workflow(self, *args, **kwargs):
        self.lazy_init_tool()
        return self._tool.create_workflow(*args, **kwargs)

    def get_task_template(self, *args, **kwargs):
        self.lazy_init_tool()
        return self._tool.get_task_template(*args, **kwargs)


def get_jobmon_tool(package) -> JobmonTool:
    return JobmonTool(package)


def make_log_dirs(log_dir: Union[str, Path]) -> Tuple[str, str]:
    """Create log directories in output root and return the paths."""
    log_dir = Path(log_dir)
    std_out = log_dir / 'output'
    std_err = log_dir / 'error'
    shell_tools.mkdir(std_out, exists_ok=True, parents=True)
    shell_tools.mkdir(std_err, exists_ok=True, parents=True)

    return str(std_out), str(std_err)


def get_cluster_name() -> 'str':
    hostname = socket.gethostname()

    if 'slurm' in hostname:
        cluster, submit_host_marker = 'slurm', 'slogin'
    elif 'uge' in hostname:
        cluster, submit_host_marker = 'buster', 'submit'
    else:
        raise RuntimeError('This tool must be run from an IHME cluster.')

    if submit_host_marker in hostname:
        raise RuntimeError("This tool must not be run from a submit host.")
    return cluster




from covid_shared.cli_tools.decorators import (
    add_verbose,
    add_with_debugger,
    add_verbose_and_with_debugger,
    pass_run_metadata,
    with_output_root,
    with_mark_best,
    with_production_tag,
    add_output_options,
    add_r_singularity_option,
    add_mobility_gpr_dependency_option,
    add_model_inputs_dependency_option,
    add_seir_covariates_dependency_option,
    add_snapshot_dependency_option,
)
from covid_shared.cli_tools.logging import (
    LOG_FORMATS,
    DEFAULT_LOG_MESSAGING_FORMAT,
    add_logging_sink,
    configure_logging_to_files,
    configure_logging_to_terminal,
)
from covid_shared.cli_tools.metadata import (
    Metadata,
    RunMetadata,
    YamlIOMixin,
    get_function_full_argument_mapping,
    monitor_application,
    handle_exceptions,
    update_with_previous_metadata,
)
from covid_shared.cli_tools.run_directory import (
    get_last_stage_directory,
    get_current_previous_version,
    get_run_directory,
    make_links,
    make_run_directory,
    mark_best,
    mark_best_explicit,
    mark_explicit,
    mark_latest,
    mark_latest_explicit,
    mark_production,
    mark_production_explicit,
    move_link,
    setup_directory_structure,
)
from covid_shared.cli_tools.cleanup import (
    finish_application,
)
from covid_shared.cli_tools.validation import (
    validate_best_and_production_tags,
)



"""Primitives for construction jobmon workflow specifications."""
import abc
import re
from typing import Dict, Type, TypeVar, Union

from loguru import logger

DEFAULT_PROJECT = 'proj_covid'
DEFAULT_QUEUE = 'd.q'

_TaskSpecDict = Dict[str, Union[str, int]]


class TaskSpecification:
    """Validation wrapper for specifying task execution parameters.

    This class is intended to outline the parameter space for task
    specification and provide sanity checks on the passed parameters so
    that we fail fast and informatively.

    Subclasses are meant to inherit and provide default values for the
    class variables. At runtime those defaults will be used to fill values
    not provided in the specification file when the specification file
    is parsed and then the full task specification will be validated against
    the values in the `validate` method of this class.

    """
    # Class variables meant to be overridden by subclasses.
    default_max_runtime_seconds: int
    default_m_mem_free: str
    default_num_cores: int

    # Constants for all tasks.
    _runtime_bounds = [100, 60 * 60 * 24]
    _mem_re = re.compile('\\d+[G]')
    _mem_bounds_gb = [1, 1000]
    _num_core_bounds = [1, 79]

    def __init__(self, task_specification_dict: _TaskSpecDict):
        self.name = self.__class__.__name__
        key_aliases_and_defaults = {
            'max_runtime_seconds': (('max_runtime_seconds', 'runtime'), self.default_max_runtime_seconds),
            'm_mem_free': (('m_mem_free', 'memory'), self.default_m_mem_free),
            'num_cores': (('num_cores', 'cores'), self.default_num_cores),
        }
        params = {}
        for key, (aliases, default) in key_aliases_and_defaults.items():
            params[key] = default
            for alias in aliases:
                if alias in task_specification_dict:
                    params[key] = task_specification_dict.pop(alias)

        self.max_runtime_seconds = task_specification_dict.pop('max_runtime_seconds',
                                                                    self.default_max_runtime_seconds)
        self.m_mem_free = task_specification_dict.pop('m_mem_free',
                                                           self.default_m_mem_free)
        self.num_cores = task_specification_dict.pop('num_cores',
                                                          self.default_num_cores)
        # Workflow specification guarantees this will be present.
        self.queue = task_specification_dict.pop('queue')

        if task_specification_dict:
            logger.warning(f'Unknown task options specified for {self.name}: {task_specification_dict}. '
                           f'These options will be ignored.')

    def __init_subclass__(cls, **kwargs):
        name = cls.__name__

        for attr in ['default_max_runtime_seconds', 'default_m_mem_free', 'default_num_cores']:
            if getattr(cls, attr) is None:
                raise AttributeError(f'No default value provided for {attr} for task template subclass {name}. '
                                     'Check your class definition.')

        if not cls._runtime_bounds[0] <= cls.default_max_runtime_seconds <= cls._runtime_bounds[1]:
            raise AttributeError(f'Invalid default max runtime for {name}: {cls.default_max_runtime_seconds}. '
                                 f'Max runtime must be in the range {cls._runtime_bounds}.'
                                 'Check your class definition.')

        if not cls._mem_re.match(cls.default_m_mem_free):
            raise AttributeError('Memory request is expected to be in the format "XG" where X is an integer. '
                                 f'You provided {cls.default_m_mem_free} for {name}.')
        if not cls._mem_bounds_gb[0] <= int(cls.default_m_mem_free[:-1]) <= cls._mem_bounds_gb[1]:
            raise AttributeError(f'Invalid default max memory for {name}: {cls.default_m_mem_free}. ',
                                 f'Max memory must be in the range {cls._mem_bounds_gb}G. '
                                 'Check your class definition.')

        if not cls._num_core_bounds[0] <= cls.default_num_cores <= cls._num_core_bounds[1]:
            raise AttributeError(f'Invalid num cores for {name}: {cls.default_num_cores}. '
                                 f'Num cores must be in the range {cls._num_core_bounds}l '
                                 'Check your class definition.')

    def validate(self):
        """Checks specification against some baseline constraints."""
        if not self._runtime_bounds[0] <= self.max_runtime_seconds <= self._runtime_bounds[1]:
            raise ValueError(f'Invalid max runtime for {self.name}: {self.max_runtime_seconds}. '
                             f'Max runtime must be in the range {self._runtime_bounds}.')

        if not self._mem_re.match(self.m_mem_free):
            raise ValueError('Memory request is expected to be in the format "XG" where X is an integer. '
                             f'You provided {self.m_mem_free} for {self.name}.')
        if not self._mem_bounds_gb[0] <= int(self.m_mem_free[:-1]) <= self._mem_bounds_gb[1]:
            raise ValueError(f'Invalid max memory for {self.name}: {self.m_mem_free}. ',
                             f'Max memory must be in the range {self._mem_bounds_gb}G.')

        if not self._num_core_bounds[0] <= self.num_cores <= self._num_core_bounds[1]:
            raise ValueError(f'Invalid num cores for {self.name}: {self.num_cores}. '
                             f'Num cores must be in the range {self._num_core_bounds}')

    def to_dict(self) -> Dict[str, Union[str, int]]:
        """Coerce the specification to a dict for display or write to disk."""
        return {'runtime': self.max_runtime_seconds,
                'memory': self.m_mem_free,
                'cores': self.num_cores,
                'queue': self.queue}

    def __repr__(self):
        return f'{self.name}({", ".join([f"{k}={v}" for k, v in self.to_dict().items()])})'


TTaskSpecification = TypeVar('TTaskSpecification', bound=TaskSpecification)


class WorkflowSpecification(abc.ABC):
    """Validation wrapper for specifying workflow execution parameters.

    This class is intended to outline the parameter space for workflow
    specification and provide sanity checks on the passed parameters so
    that we fail fast and informatively.

    Subclasses are meant to inherit and provide a dict mapping task type
    names to concrete subclasses of the ``TaskSpecification`` base class
    as the ``tasks`` class variable.  When instantiated, the
    ``WorkflowSpecification`` subclasses will automatically resolve provided
    specification values with defaults and do some sanity check validation.

    """

    tasks: Dict[str, Type[TTaskSpecification]]

    def __init__(self,
                 tasks: Dict[str, Dict[str, Union[int, str]]] = None,
                 project: str = None,
                 queue: str = None):
        self.name: str = self.__class__.__name__
        self.project: str = project if project is not None else DEFAULT_PROJECT
        self.queue: str = queue if queue is not None else DEFAULT_QUEUE

        # Check everything's okay before making the task specs
        self.validate()
        tasks = {} if tasks is None else tasks
        self.task_specifications: Dict[str, TaskSpecification] = self.process_task_dicts(tasks)

    def validate(self):
        """Checks specification against some baseline constraints."""
        allowed_projects = ['proj_dq', 'proj_covid', 'proj_covid_prod']
        if self.project not in allowed_projects:
            raise ValueError(f'Invalid project selected for {self.name}: {self.project}. '
                             f'Covid workflows should be run on one of {allowed_projects}.')
        allowed_queues = ['d.q', 'all.q', 'long.q']
        if self.queue not in allowed_queues:
            raise ValueError(f'Invalid queue for {self.name}: {self.queue}. '
                             f'Queue must be one of {allowed_queues}.')

    def process_task_dicts(self,
                           task_specification_dicts: Dict[str, _TaskSpecDict]) -> Dict[str, TaskSpecification]:
        """Converts parsed input specifications for tasks into concrete
        instances of ``TaskSpecification`` subclasses.

        """
        task_specifications = {}
        for task_name, spec_class in self.tasks.items():
            task_spec_dict = task_specification_dicts.pop(task_name, {})
            task_spec_dict['queue'] = self.queue
            task_specifications[task_name] = spec_class(task_spec_dict)

        if task_specification_dicts:
            logger.warning(f'Task specifications provided for unknown tasks: {list(task_specification_dicts)}.'
                           f'These specifications will be ignored.')

        return task_specifications

    def to_dict(self) -> Dict:
        """Coerce the specification to a dict for display or write to disk."""
        return {'project': self.project,
                'queue': self.queue,
                'tasks': {k: v.to_dict() for k, v in self.task_specifications.items()}}

    def __repr__(self):
        return f'{self.name}({", ".join([f"{k}={v}" for k, v in self.to_dict().items()])})'



import pytest

from covid_shared.ihme_deps import _lazy_import_callable


def test_lazy_import_callable_pass():
    from itertools import chain
    chain2 = _lazy_import_callable('itertools', 'chain')
    assert chain is chain2


def test_lazy_import_callable_fail():
    magic = _lazy_import_callable('fairyland', 'magic')
    with pytest.raises(ModuleNotFoundError):
        magic()
